# Test of insert_head, insert_tail, reverse, and remove_head
option fail 0
option malloc 0
new
ih dolphin
ih bear
ih gerbil
reverse
it meerkat
it bear
it gerbil
reverse
it squirrel
reverse
rh squirrel
ih vulture
reverse
rh gerbil
rh bear
rh meerkat
rh gerbil
rh bear
rh dolphin
rh vulture
