# Test of insert_head, insert_tail, remove_head reverse, and size
option fail 0
option malloc 0
new
ih dolphin
ih bear
ih gerbil
reverse
size
it meerkat
it bear
it gerbil
size
rh dolphin
reverse
size
rh gerbil
rh bear
rh meerkat
rh gerbil
rh bear
size
free
