# Test of insert_head, insert_tail, and remove_head
option fail 0
option malloc 0
new
ih gerbil
ih bear
ih dolphin
it meerkat
it bear
it gerbil
rh dolphin
rh bear
rh gerbil
rh meerkat
rh bear
rh gerbil
